import os
from tempfile import TemporaryDirectory
import itertools
import importlib
import logging

from .transfer import get_downloader, get_uploader
from .archive import unpack_files
from .metadata import extract_product_type_and_level


logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

def custom_preprocessor(source_dir, target_dir, path, args=None, kwargs=None):
    """ Preprocessing step for a custom preprocessing.
    """
    module_name, _, func_name = path.rpartition('.')
    func = getattr(importlib.import_module(module_name), func_name)
    func(source_dir, target_dir, *(args or []), **(kwargs or {}))


def georeference(source_dir: os.PathLike, target_dir: os.PathLike, type: str):
    """ Apply georeferencing to the files in source_dir to create georeferenced
        image files in target_dir.
        Georeferencing type specifies what geoereferencing shall be applied. Options
        are 'GCP', 'RPS', 'world', ...
    """
    # TODO: implement
    pass


def stack_bands(source_dir, target_dir, group_by, file_order):
    """ Stack bands of the individual images
    """
    # TODO: implement
    pass

def output(source_dir, target_dir, crs, driver, format_options):
    pass


def custom_postprocessor(source_dir, target_dir, path, args=None, kwargs=None):
    """ Preprocessing step for a custom preprocessing.
    """
    module_name, _, func_name = path.rpartition('.')
    func = getattr(importlib.import_module(module_name), func_name)
    func(source_dir, target_dir, *(args or []), **(kwargs or {}))


STEP_FUNCTIONS = {
    'custom_preprocessor': custom_preprocessor,
    'georeference': georeference,
    'stack_bands': stack_bands,
    'output': output,
    'custom_postprocessor': custom_postprocessor,
}

def flatten(l):
    return [item for sublist in l for item in sublist]


# -----------------------------------------------------------------------------

def preprocess_file(config: dict, file_path: os.PathLike):
    """ Runs the preprocessing of a single file.
    """
    with TemporaryDirectory(prefix='preprocess_') as tempdir:
        os.chdir(tempdir)
        os.mkdir('download')
        os.mkdir('unpack')

        # get the Downloader for the configured source archive to download the given source file
        source_config = config['source']
        downloader = get_downloader(
            source_config['type'], source_config.get('args'), source_config.get('kwargs')
        )
        source_archive_path = downloader.download(file_path, 'download')

        # fetch the metadata XML file from the downloaded archive
        metadata_files = unpack_files(source_archive_path, 'unpack', glob=config['metadata_glob'])

        # open the XML to retrieve the product type and level
        product_type, product_level = extract_product_type_and_level(metadata_files, config)

        # get a concrete configuration for the type, filled with the defaults
        default_config = dict(config['preprocessing'].get('defaults', {}))
        preprocess_config = dict(config['preprocessing']['types'].get(product_type, {}))
        default_config.update(preprocess_config)

        # select and unpack files according to configuration
        data_files = flatten([
            unpack_files(
                source_archive_path,
                'unpack',
                glob=glob,
                recursive=preprocess_config.get('nested', False),
            )
            for glob in preprocess_config['data_file_globs']
        ])

        additional_files = flatten([
            unpack_files(
                source_archive_path,
                'unpack',
                glob=glob,
                recursive=preprocess_config.get('nested', False),
            )
            for glob in preprocess_config['additional_file_globs']
        ])

        previous_step = 'unpack'
        # make processing steps
        for step in ['custom_preprocessor', 'georeference', 'stack_bands', 'output', 'custom_postprocessor']:
            step_config = preprocess_config.get(step)
            if not step_config:
                continue

            os.mkdir(step)

            preprpocessor = STEP_FUNCTIONS[step]
            preprpocessor(previous_step, step, **step_config)

            previous_step = step


        # get an uploader for the finalized images
        target_config = config['target']
        uploader = get_uploader(
            target_config['type'], target_config.get('args'), target_config.get('kwargs')
        )

        # TODO send all files in the output directory to the target storage
        uploader.upload()